
'''Classes that handle proxying for RTP Sessions and provide a command port
to receive commands from SER (SIP Express Router).

Copyright 2003-2004 Dan Pascu
'''

import asyncore
import socket
import sys, os, grp, errno, traceback, atexit
import time
import struct
import re
import weakref


## Non editable settings (constants or parameters overwritten by command line options)

controlSock      = '/var/run/mediaproxy.sock'
controlSockGroup = 'ser'
tcpListenAddress = ('0.0.0.0', 25060)

idleTimeout =    60 ## Session timeout if no data is received (in seconds)
holdTimeout = 60*60 ## Session timeout if no data is received when on hold (in seconds)
forceClose  =     0 ## Close a sesion after this much time (0 for no timeout)

proxyIP = '0.0.0.0'
minPort = 35000     ## Must be even!!!
maxPort = 65000     ## Must be even!!!
crtPort = minPort   ## Must be even!!!
TOS = 0xb8

## Internal state of mediaproxy (DON'T TOUCH!!!)

MediaProxyVersion = "1.2.1"

Sessions = {}

## For computing traffic through mediaproxy
byteBucket = [0, 0, 0]
Traffic    = [0, 0, 0]

## Non public networks. This includes RFC1918 networks, localnet (127.0.0.0/8),
## broadcast networks (224.0.0.0/4) and 0.0.0.0/8
nonPublicNetworks = [
    {'name': '0.0.0.0',     'value': 0x00000000L, 'mask': 0xff000000L},
    {'name': '10.0.0.0',    'value': 0x0a000000L, 'mask': 0xff000000L},
    {'name': '127.0.0.0',   'value': 0x7f000000L, 'mask': 0xff000000L},
    {'name': '172.16.0.0',  'value': 0xac100000L, 'mask': 0xfff00000L},
    {'name': '192.168.0.0', 'value': 0xc0a80000L, 'mask': 0xffff0000L},
    {'name': '224.0.0.0',   'value': 0xe0000000L, 'mask': 0xf0000000L}
]

rtpPayloads = {
     0: 'G711u', 1: '1016',  2: 'G721',  3: 'GSM',  4: 'G723',  5: 'DVI4', 6: 'DVI4',
     7: 'LPC',   8: 'G711a', 9: 'G722', 10: 'L16', 11: 'L16',  14: 'MPA', 15: 'G728',
    18: 'G729', 25: 'CelB', 26: 'JPEG', 28: 'nv',  31: 'H261', 32: 'MPV', 33: 'MP2T',
    34: 'H263'
}

## Log network traffic into mysql database if dblog.py exists

try:
    from dblog import *
except ImportError:
    radiusdb = None
else:
    try:
        import dbcfg
    except ImportError:
        radiusdb = None
    else:
        del dbcfg
        try: 
            radiusdb = RadiusAccounting()
        except RuntimeError, why:
            print >>sys.stderr, "MySQL warning: '%s'. Will not log traffic to database." % why
            radiusdb = None

## Helper functions

def warning(msg):
    print >>sys.stderr, "warning: " + str(msg)

def error(msg):
    print >>sys.stderr, "error: " + str(msg)

def getHostIP():
    '''Will return the host's main IP address'''
    try:
        return socket.gethostbyname(socket.getfqdn())
    except:
        return None

def isPublicIP(address):
    try:
        adr = socket.inet_aton(address)
    except:
        if address:
            print >>sys.stderr, "Invalid IP address: '%s'" % address
        return False
    for net in nonPublicNetworks:
        netaddr = struct.unpack('!L', adr)[0] & net['mask']
        if (netaddr == net['value']):
            return False
    return True

def getNetRange(description):
    '''
    Takes the input network[/mask] entry and returns a tuple consisting of
    the network base address and the network mask.
    Input should be one of:
        - network/mask
        - ip_address
        - hostname
    in the latter two cases a mask of 32 is assumed.
    Except the hostname case, where a DNS name is present, in the other cases
    the address should be in quad dotted number notation.
    Mask is number between 0 and 32 (bits used by the network part of address)
    On error ValueError is raised.
    '''
    if not description:
        return (0L, 0xFFFFFFFFL)
    if description.lower()=='any':
        return (0L, 0L) ## This is the any address 0.0.0.0
    match = re.search(r'(?P<net>.+?)/(?P<bits>\d+)', description)
    if match:
        net     = match.group('net')
        netbits = int(match.group('bits'))
    else:
        try:
            net = socket.gethostbyname(description) # if not a net/mask it may be a host or ip
        except socket.gaierror:
            raise NameError, "Invalid hostname or IP in address: '%s'" % description
        netbits = 32
    if netbits < 0 or netbits > 32:
        raise ValueError, "Invalid network mask in address: '%s'" % description
    try:
        netaddr = socket.inet_aton(net)
    except:
        raise ValueError, "Invalid IP in address: '%s'" % net
    mask = (0xFFFFFFFFL << 32-netbits) & 0xFFFFFFFFL
    #mask = (-1L << 32-netbits) & 0xFFFFFFFFL
    #mask = ((1L << netbits) - 1) << 32-netbits
    netbase = struct.unpack('!L', netaddr)[0] & mask
    return netbase, mask


## We overwrite the loop in the asyncore module to be able to call an extra function
## at the end of the loop (to let us timeout idle RTP sessions).
## While timeout can be a fractional number, we only use integers.
def loop(timeout=1, map=None):
    if map is None:
        map = asyncore.socket_map
    timer = SessionTimer(timeout)
    while map:
        asyncore.poll3(timeout, map)
        timeout = timer.run()

## Classes

class CommandError(Exception):     pass
class MediaProxyError(Exception):  pass
class RTPSessionError(Exception):  pass
class MediaStreamError(Exception): pass
class RTPStreamError(Exception):   pass


class SessionTimer:
    def __init__(self, period):
        self.start = time.time()
        self.period = period
        self.next = self.start + period
        #self.before = self.start

    def run(self):
        global byteBucket, Traffic
        now = time.time()
        if now >= self.next:
            period = self.period
            ## Check sessions and timeout idle ones
            for session in Sessions.values():
                for stream in session.mediaStreams:
                    stream.idleTime += period
                if session.didTimeout:
                    session.end()
                elif forceClose and session.duration >= forceClose:
                    session.forceClosed = True
                    session.end()
            Traffic    = [int(val/period) for val in byteBucket]
            byteBucket = [0, 0, 0]
            self.next += period
            #print "Loop time: %5.3f ms" % ((time.time() - now)*1000)
            return self.next - time.time()
        return self.next - now


class ControlSocket(asyncore.dispatcher):
    '''Control MediaProxy through a local socket on the filesystem'''
    def __init__(self, path=None, group=None):
        ## If we would have passed these as default values they would have been
        ## locked to the values they had when the rtphandler module is loaded
        ## This way we can change the default values after rtphandler is loaded
        ## and we get the updated values
        path = path or controlSock
        group = group or controlSockGroup
        self.path = path
        try:
            os.unlink(path)
        except OSError:
            pass
        asyncore.dispatcher.__init__(self)
        self.create_socket(socket.AF_UNIX, socket.SOCK_STREAM)
        ## we don't need this for unix sockets #self.set_reuse_addr()
        try:
            self.bind(path)
            self.listen(5)
        except socket.error, why:
            raise MediaProxyError, "couldn't create command socket (%s)" % why[1]
        try:
            groupId = grp.getgrnam(group)[2]
        except KeyError, IndexError:
            ## Make it world writable
            try:
                os.chmod(path, 0777)
            except OSError:
                warning("couldn't set access rights for %s." % path)
                warning("the SIP proxy may be unable to communicate with us!")
        else:
            ## Make it writable only to the SIP proxy group members
            try:
                os.chown(path, -1, groupId)
                os.chmod(path, 0770)
                #os.chmod(path, S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IWGRP|S_IXGRP)
            except OSError:
                warning("couldn't set access rights for %s." % path)
                warning("the SIP proxy may be unable to communicate with us!")
        atexit.register(self.handle_close)

    def writable(self):
        return 0

    def handle_accept(self):
        try:
            connection, address = self.accept()
        except socket.error: ## Rare Linux error
            print >>sys.stderr, "Error accepting connection on control socket"
            return
        except TypeError: ## Rare FreeBSD3 error
            print >>sys.stderr, "EWOULDBLOCK exception on control socket"
            return
        cmdHandler = CommandHandler(connection, address)

    def handle_close(self):
        self.close()
        try:
            os.unlink(self.path)
        except OSError:
            pass


class RemoteControl(asyncore.dispatcher):
    '''Control MediaProxy from remote through a TCP socket'''
    def __init__(self, address=None, allowedHosts=()):
        address = address or tcpListenAddress
        asyncore.dispatcher.__init__(self)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr()
        try:
            self.bind(address)
            self.listen(5)
        except socket.error, why:
            raise MediaProxyError, "couldn't create command socket (%s)" % why[1]
        allowed = []
        for net in allowedHosts:
            try:
                allowed.append(getNetRange(net))
            except NameError:
                print "Couldn't resolve hostname: `%s' (ignored)" % net
            except ValueError:
                print "Invalid network specification: `%s' (ignored)" % net
        if (0L, 0L) in allowed:
            self.allowedHostList = ()
        else:
            self.allowedHostList = allowed
        atexit.register(self.handle_close)

    def isAllowed(self, address):
        if not self.allowedHostList:
            return True
        try:
            adr = socket.inet_aton(address)
        except:
            return False
        for netbase, mask in self.allowedHostList:
            if (struct.unpack('!L', adr)[0] & mask) == netbase:
                return True
        return False

    def writable(self):
        return 0

    def handle_accept(self):
        try:
            connection, address = self.accept()
        except socket.error: ## Rare Linux error
            print >>sys.stderr, "Error accepting connection on control socket"
            return
        except TypeError: ## Rare FreeBSD3 error
            print >>sys.stderr, "EWOULDBLOCK exception on control socket"
            return
        if not self.isAllowed(address[0]):
            connection.close()
            print >>sys.stderr, "Connection from %s is not allowed" % address[0]
            return
        cmdHandler = CommandHandler(connection, address)

    def handle_close(self):
        self.close()


class CommandHandler(asyncore.dispatcher):
    '''Handles a command received through the control socket'''

    commandMap = {'request': {'minArguments': 9},
                  'lookup':  {'minArguments': 9},
                  'delete':  {'minArguments': 1},
                  'status':  {'minArguments': 0},
                  'status2': {'minArguments': 0},
                  'summary': {'minArguments': 0},
                  'version': {'minArguments': 0}}
    
    def __init__(self, connection, address):
        asyncore.dispatcher.__init__(self, connection)
        self.addr = address
        self.outBuffer = ''
        self.inBuffer = ''

    def writable(self):
        return len(self.outBuffer) > 0

    def handle_read(self):
        command = self.recv(1024)
        if not command:
            self.close()
            self.inBuffer  = ''
            self.outBuffer = ''
            return
        try:
            command.index('\n')
        except ValueError:
            self.inBuffer += command
            if len(self.inBuffer) > 8192:
                ## Somebody is flooding
                self.close()
                self.inBuffer = ''
                self.outBuffer = ''
            return
        else:
            command = self.inBuffer + command
            self.inBuffer = ''
        start = time.time()
        try:
            result = self.handleCommand(command)
        except CommandError, msg:
            error(msg)
            self.close()
        else:
            if result is None:
                self.close()
            else:
                self.outBuffer = "%s\n" % result
        print "command execution time: %5.2f ms" % ((time.time() - start) * 1000)

    def handle_write(self):
        sent = self.send(self.outBuffer)
        self.outBuffer = self.outBuffer[sent:]
        if not self.outBuffer:
            self.close()

    def handle_close(self):
        self.close()

    def handleCommand(self, command):
        global Sessions

        command = command.rstrip()
        print "command %s" % command

        ## Split command and check if it is valid
        sessionData = command.split()
        if len(sessionData) < 1:
            raise CommandError, "received empty command '%s'" % command
        cmd = sessionData[0].lower()
        if cmd not in self.commandMap.keys():
            raise CommandError, "received unknown command '%s'" % command

        minArguments = self.commandMap[cmd]['minArguments']
        if len(sessionData) < minArguments + 1:
            suffix = (minArguments!=1 and 's') or ''
            raise CommandError, "the '%s' command needs at least %d argument%s: '%s'" % \
                  (cmd, minArguments, suffix, command)

        ## And do the mumbo-jumbo thing
        if cmd == 'status':
            result = ["proxy %s %s/%s/%s" % ((len(Sessions),) + tuple(Traffic))]
            for session in Sessions.values():
                result.append(session.statline)
                for stream in session.mediaStreams:
                    result.append(stream.statline)
            return '\n'.join(result)
        elif cmd == 'status2':
            result = ["proxy %s %s/%s/%s" % ((len(Sessions),) + tuple(Traffic))]
            for session in Sessions.values():
                for stream in session.mediaStreams:
                    rtp = stream.rtpStream
                    idleTime = stream.idleTime
                    try:    adr1 = rtp.caller.paddress
                    except: adr1 = '?.?.?.?:?'
                    try:    adr2 = rtp.called.paddress
                    except: adr2 = '?.?.?.?:?'
                    ## We test for hold first because a call can be put on hold
                    ## before RTP data is sent (if phones have silence detection
                    ## capabilities for example)
                    if session.onHold:     status = 'hold'
                    elif rtp.signedIn < 2: status = 'inactive'
                    elif idleTime > 2:     status = 'idle'
                    else:                  status = 'active'
                    pbytes = "%s/%s/%s" % stream.bytes
                    line = '%s %s %s %s %s %s %s %s %s %s %s %s %s %s' % \
                         (session._id, adr1, rtp.paddr, adr2, status,
                          session.duration, idleTime, pbytes, rtp.rtpCodec,
                          stream.type,
                          session.callerInfo['from'], session.callerInfo['to'],
                          session.encodedAgents[0], session.encodedAgents[1])
                    result.append(line)
            return '\n'.join(result)
        elif cmd == 'summary':
            return "%s %s/%s/%s" % ((len(Sessions),) + tuple(Traffic))
        elif cmd == 'version':
            return MediaProxyVersion
        elif cmd == 'request':
            sessionId = sessionData[1]
            try:
                session = Sessions[sessionId]
            except KeyError:
                ## Create a brand new session
                try:
                    session = RTPSession(sessionData)
                except RTPSessionError, why:
                    raise CommandError, "cannot create session for '%s': %s" % (sessionId, why)
            else:
                session.updateStreams(sessionData)
            ips = [address.split(':')[0] for address in sessionData[2].split(',')]
            ## A session can be put on hold only after both parties have sent their
            ## contact data. However they don't have to sign in on RTP before
            ## they can put the call on hold (they may have silence detection)
            if session.complete and '0.0.0.0' in ips:
                session.onHold  = True
                session.timeout = holdTimeout
            else:
                session.onHold  = False
                session.timeout = idleTimeout
            return session.endpointAddresses()
        elif cmd == 'lookup':
            sessionId = sessionData[1]
            try:
                session = Sessions[sessionId]
            except KeyError:
                warning("trying to lookup session with non-existent id: '%s'" % sessionId)
                return None
            session.fillInCalledParty(sessionData)
            # remove (or make it show in debug mode only)
            #for stream in session.mediaStreams:
                #stream.rtpStream.caller.show('rtp', compact=1)
                #stream.rtcpStream.caller.show('rtcp', compact=1)
                #stream.rtpStream.called.show('rtp', compact=1)
                #stream.rtcpStream.called.show('rtcp', compact=1)
            # until here
            return session.endpointAddresses()
        elif cmd == 'delete':
            sessionId = sessionData[1]
            try:
                session = Sessions[sessionId]
            except KeyError:
                pass
            else:
                session.end()
            return None
        else:
            raise CommandError, "received unknown command: '%s'" % command.rstrip()


class RTPPeer(object):
    '''A RTPPeer represents one endpoint of a RTP stream. There will be 4 RTPPeer's
    in a RTP session (2 from the data stream + 2 from the control stream)'''

    class Values:
        '''Hold the actual parameter values for a RTPPeer'''
        def __init__(self):
            self._parent     = None
            self.contactIP   = None
            self.visibleIP   = None
            self.expectedIP  = None
            self.rtpPort     = 0
            self.asymmetric  = 0
            self.local       = 0
            self.address     = None
            self.paddress    = '?.?.?.?:?' ## printable form of self.address
            self.signedInAddress = None
            self.expectedAddress = None
        def updateContacts(self):
            signedInIP   = (self.signedInAddress and self.signedInAddress[0]) or None
            signedInPort = (self.signedInAddress and self.signedInAddress[1]) or 0
            publicContactIP = (isPublicIP(self.contactIP) and self.contactIP) or None
            probableIncomingIP = (self.local and self.visibleIP) or None
            # This is problematic. contactIP may be a public IP, yet still different
            # from visibleIP (if the client uses the address of a conference server
            # for example)
            #if self.local and publicContactIP and self.visibleIP != self.contactIP:
                ### Client is using public IP address behind NAT. Ignore.
                #publicContactIP = None
            ## add probableIncomingIP to contactIP if you want to start sending to a peer
            ## that is behind NAT before he signs in. This requires static port forwarding
            ## in the NAT box for asymmetric clients to work, while for symmetric clients
            ## the chances to reach them are dim and depend on the type of NAT and if the
            ## connection was before opened from inside and have not yet timed out.
            contactIP = signedInIP or publicContactIP #or probableIncomingIP
            if self.asymmetric:
                port = self.rtpPort
            else:
                port = signedInPort or self.rtpPort
            self.address = (contactIP and port and (contactIP, port)) or None
            self.paddress = "%s:%s" % (self.address or ('?.?.?.?', '?'))
            self.expectedIP = publicContactIP or probableIncomingIP
            self.expectedAddress = (self.expectedIP and port and (self.expectedIP, port)) or None

    def __init__(self, name, contactIP=None, rtpPort=0, visibleIP=None, local=0, asymmetric=0):
        self.__v = self.Values()
        self.__v._parent = weakref.proxy(self)
        self.__v.contactIP = contactIP
        self.__v.visibleIP = visibleIP
        self.__v.rtpPort = rtpPort
        self.__v.asymmetric = asymmetric
        self.__v.local = local
        self.__v.signedInAddress = None
        self.__v.updateContacts()
        self.name = name

    def getci(self): return self.__v.contactIP
    def getvi(self): return self.__v.visibleIP
    def getei(self): return self.__v.expectedIP
    def getea(self): return self.__v.expectedAddress
    def getrp(self): return self.__v.rtpPort
    def getas(self): return self.__v.asymmetric
    def getlo(self): return self.__v.local
    def getca(self): return self.__v.address
    def getpa(self): return self.__v.paddress
    def getsa(self): return self.__v.signedInAddress
    def setci(self, val): self.__v.contactIP       = val; self.__v.updateContacts()
    def setvi(self, val): self.__v.visibleIP       = val; self.__v.updateContacts()
    def setrp(self, val): self.__v.rtpPort         = val; self.__v.updateContacts()
    def setas(self, val): self.__v.asymmetric      = val; self.__v.updateContacts()
    def setlo(self, val): self.__v.local           = val; self.__v.updateContacts()
    def setsa(self, val): self.__v.signedInAddress = val; self.__v.updateContacts()
    contactIP  = property(getci, setci)
    visibleIP  = property(getvi, setvi)
    expectedIP = property(getei)
    rtpPort    = property(getrp, setrp)
    asymmetric = property(getas, setas)
    local      = property(getlo, setlo)
    address    = property(getca)
    paddress   = property(getpa)  ## Printable form of self.address
    signedInAddress = property(getsa, setsa)
    expectedAddress = property(getea)
    del(getci, getvi, getei, getea, getrp, getas, getca, getpa, getsa, getlo)
    del(setci, setvi, setrp, setas, setsa, setlo)

    def show(self, type, compact=0):
        if compact:
            print "%s.%-4s: sendto=%s expectIP=%s port=%d local=%d asym=%d" % (
                self.name, type.lower(), str(self.address), self.expectedIP,
                self.rtpPort, self.local, self.asymmetric)
        else:
            print "Contact IP:  %s" % self.contactIP
            print "Visible IP:  %s" % self.visibleIP
            print "Expected IP: %s" % self.expectedIP
            print "Signed in A: %s" % str(self.signedInAddress)
            print "Contact Adr: %s" % str(self.address)
            print "Port       : %d" % self.rtpPort
            print "Local      : %d" % self.local
            print "Asymmetric : %d" % self.asymmetric


class RTPStream(asyncore.dispatcher):
    '''A RTPStream is an entity consisting in a RTP proxy and 2 RTP endpoints (RTPPeers).
    The RTP proxy is relaying the data between the 2 RTP endpoints both ways.
    The RTP proxy is intrinsic in the structure of the RTPStream class, while the RTPPeer's
    are implemented by an external class.'''

    def __init__(self, stream, sock, name='RTP'):
        asyncore.dispatcher.__init__(self, sock)
        self.addr = sock.getsockname()
        self.paddr = "%s:%s" % self.addr
        self.stream = weakref.proxy(stream) ## The media stream we belong to
        self.name = name
        self.caller = None
        self.called = None
        self.bytes   = [0, 0, 0]  ## from caller, from called, relayed
        self.packets = [0, 0, 0]  ## from caller, from called, relayed
        self.payload = None
        self.rtpCodec = 'Unknown'
        self.signedIn = 0
        self.warned3rdparty = 0

    def recvfrom(self, buffer_size):
        try:
            return self.socket.recvfrom(buffer_size)
        except socket.error, why:
            if why[0] != errno.EWOULDBLOCK:
                self.handle_close()
                raise socket.error, why
            return '', ('dummy', 0)

    def sendto(self, data, address):
        try:
            return self.socket.sendto(data, address)
        except socket.error, why:
            if why[0] != errno.EWOULDBLOCK:
                self.handle_close()
                raise socket.error, why
            return 0

    def writable(self):
        return 0

    def setCodec(self, data):
        try:
            payload = ord(data[1]) & 0x7f
        except IndexError:
            print >>sys.stderr, "warning: insufficient data in RTP packet to determine codec type"
        else:
            ## payload=13  is 'Confort noise'
            ## payload=101 is 'Telephone event'
            if payload not in (13, 101):
                self.payload = payload
                unknownName = ((payload>95 and 'Dynamic(%d)') or 'Unknown(%d)') % payload
                self.rtpCodec = rtpPayloads.get(payload, unknownName)

    def signIn(self, peer, address, data):
        peer.signedInAddress = address
        self.signedIn += 1
        self.setCodec(data)
        # eventually remove (or make it show only when debug is enabled)
        print "session %s: %s signed in from %s:%s (%s) (will return to %s:%s)" % \
              ((self.stream.session._id, peer.name) + address + (self.name,) + peer.address)

    def handle_read(self):
        global byteBucket

        data, address = self.recvfrom(65536)
        if not (data and self.stream.complete):
            return
        bytes = len(data) + 28 ## 28 is from IP+UDP headers
        caller = self.caller
        called = self.called
        if address == caller.signedInAddress:
            sender = 0
            destination = called.address
        elif address == called.signedInAddress:
            sender = 1
            destination = caller.address
        elif not caller.signedInAddress and address == caller.expectedAddress:
            self.signIn(caller, address, data)
            sender = 0
            destination = called.address
        elif not called.signedInAddress and address == called.expectedAddress:
            self.signIn(called, address, data)
            sender = 1
            destination = caller.address
        elif not caller.signedInAddress and address[0] == caller.expectedIP:
            self.signIn(caller, address, data)
            sender = 0
            destination = called.address
        elif not called.signedInAddress and address[0] == called.expectedIP:
            self.signIn(called, address, data)
            sender = 1
            destination = caller.address
        elif not caller.signedInAddress and caller.local and address[0]==caller.visibleIP:
            self.signIn(caller, address, data)
            sender = 0
            destination = called.address
        elif not called.signedInAddress and called.local and address[0]==called.visibleIP:
            self.signIn(called, address, data)
            sender = 1
            destination = caller.address
        elif not caller.signedInAddress and not caller.expectedIP:
            self.signIn(caller, address, data)
            sender = 0
            destination = called.address
        elif not called.signedInAddress and not called.expectedIP:
            self.signIn(called, address, data)
            sender = 1
            destination = caller.address
        else:
            if not self.warned3rdparty:
                self.warned3rdparty = 1
                warning("Received packet from a third party: %s:%s" % address)
            return
        if self.payload is None:
            self.setCodec(data)
        byteBucket[sender]   += bytes
        self.bytes[sender]   += bytes
        self.packets[sender] += 1
        if destination:
            byteBucket[2]    += bytes
            self.bytes[2]    += bytes
            self.packets[2]  += 1
            self.sendto(data, destination)
        self.stream.idleTime = 0

    def handle_close(self):
        self.close()


class MediaStream(object):
    def __init__(self, session, mediatype='Audio'):
        self.session = weakref.proxy(session)
        self.type = mediatype.capitalize()
        (dataSlot, ctrlSlot) = self.__findRTPSlot()
        self.rtpStream  = RTPStream(self, dataSlot, 'RTP')
        self.rtcpStream = RTPStream(self, ctrlSlot, 'RTCP')
        self.idleTime = 0
        self.complete = 0

    ## complete is when the media stream is described for both ends
    #def getco(self): return (self.rtpStream.caller and self.rtpStream.called)
    #complete = property(getco)
    #del(getco)

    def getby(self):
        di, do, dr = self.rtpStream.bytes
        ci, co, cr = self.rtcpStream.bytes
        return (di+ci, do+co, dr+cr)
    def getpa(self):
        di, do, dr = self.rtpStream.packets
        ci, co, cr = self.rtcpStream.packets
        return (di+ci, do+co, dr+cr)
    def getsl(self):
        rtp = self.rtpStream
        caller = rtp.caller
        called = rtp.called
        idleTime = self.idleTime
        try:    adr1 = caller.paddress
        except: adr1 = '?.?.?.?:?'
        try:    adr2 = called.paddress
        except: adr2 = '?.?.?.?:?'
        ## We test for hold first because a call can be put on hold before RTP
        ## data is sent (if phones have silence detection capabilities for example)
        if self.session.onHold: status = 'hold'
        elif rtp.signedIn < 2:  status = 'inactive'
        elif idleTime > 2:      status = 'idle'
        else:                   status = 'active'
        pbytes = "%s/%s/%s" % self.bytes
        return 'stream %s %s %s %s %s %s %s %s' % \
               (adr1, adr2, rtp.paddr, pbytes, status, rtp.rtpCodec,
                self.type, idleTime)

    bytes    = property(getby)
    packets  = property(getpa)
    statline = property(getsl)
    
    del(getby, getpa, getsl)
    
    def close(self):
        self.rtpStream.close()
        self.rtcpStream.close()

    def setCaller(self, address, visibleIP, asymmetric):
        if not self.rtpStream.caller:
            ip = address[0]
            port = int(address[1] or 0)
            local = self.session.localCaller
            rtpStream  = self.rtpStream
            rtcpStream = self.rtcpStream
            rtpStream.caller  = RTPPeer('caller', ip, port,   visibleIP, local, asymmetric)
            rtcpStream.caller = RTPPeer('caller', ip, port+1, visibleIP, local, asymmetric)
            self.complete = rtpStream.called is not None

    def setCalled(self, address, visibleIP, asymmetric):
        if not self.rtpStream.called:
            ip = address[0]
            port = int(address[1] or 0)
            local = self.session.localCalled
            rtpStream  = self.rtpStream
            rtcpStream = self.rtcpStream
            rtpStream.called  = RTPPeer('called', ip, port,   visibleIP, local, asymmetric)
            rtcpStream.called = RTPPeer('called', ip, port+1, visibleIP, local, asymmetric)
            self.complete = rtpStream.caller is not None

    def __findRTPSlot(self):
        '''Find a pair of consecutive ports to bind to and return
        two sockets bound tho them. The first port must be even.'''

        def validateRTPPort(port):
            if (port & 0x01): port += 1
            if port < minPort or port >= maxPort: port = minPort
            return port

        global crtPort

        dsock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        csock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            dsock.setsockopt(socket.IPPROTO_IP, socket.IP_TOS, TOS)
            csock.setsockopt(socket.IPPROTO_IP, socket.IP_TOS, TOS)
        except socket.error:
            pass ## Not critical
        crtPort = validateRTPPort(crtPort)
        for port in range(crtPort, maxPort, 2) + range(minPort, crtPort, 2):
            try:
                try:
                    dsock.bind((proxyIP, port))
                except socket.error, why:
                    if why[0] in (errno.EADDRINUSE, errno.EACCES):
                        continue
                    else:
                        raise
                else:
                    try:
                        csock.bind((proxyIP, port+1))
                    except socket.error, why:
                        if why[0] in (errno.EADDRINUSE, errno.EACCES):
                            dsock.close()
                            dsock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                            continue
                        else:
                            raise
                    else:
                        crtPort = port + 2
                        return (dsock, csock)
            except:
                ## Print traceback
                info = sys.exc_info()
                try:
                    msg = ''.join(traceback.format_exception(*info))
                    print >>sys.stderr, msg,
                finally:
                    del info
                break
        dsock.close()
        csock.close()
        raise MediaStreamError, "cannot find a free pair of ports for a media stream"


class RTPSession(object):
    def __init__(self, sessionData):
        global Sessions

        self._id = sessionData[1]
        self.callerDomain = sessionData[4]
        self.calledDomain = sessionData[6]
        self.localCaller  = (sessionData[5] == 'local')
        self.localCalled  = (sessionData[7] == 'local')
        self.userAgents = [sessionData[8].decode('quopri'), 'Unknown']
        self.encodedAgents = [sessionData[8], 'Unknown']
        visibleIP = sessionData[3]
        info = self.parseInfo(sessionData[9])
        asymmetric = info.get('asymmetric', 0)
        self.callerInfo = info
        ## We need this info until the called party signs in, else if a CANCEL
        ## from the caller comes before the OK it will raise an exception because
        ## there is no calledInfo to compute the stats when the session ends
        self.calledInfo = {'from':    info['from'],    'to':    info['to'],
                           'fromtag': info['fromtag'], 'totag': info['totag']}
        streams = [s.split(':') for s in sessionData[2].split(',')]
        self.mediaStreams = []
        for sinfo in streams:
            try:
                mediatype = sinfo[2]
            except IndexError:
                mediatype = 'Audio'
                # remove at a later time
                print "warning: SER is using an old mediaproxy.so module. " + \
                      "Please upgrade it to get correct media type information."
            try:
                stream = MediaStream(self, mediatype)
            except MediaStreamError:
                for stream in self.mediaStreams:
                    stream.close()
                raise RTPSessionError, "cannot find enough free ports for the media streams"
            stream.setCaller(sinfo, visibleIP, asymmetric)
            self.mediaStreams.append(stream)
        if not self.mediaStreams:
            raise RTPSessionError, "there are no media streams to process"
        self.startTime = round(time.time())
        self.conversationStart = None
        self.timeout = idleTimeout
        self.onHold = False
        self.forceClosed = False
        self.complete = 0 ## Will need the destination filled in as well
        self.needStreamUpdate = 0 ## If new streams are added on the fly

        Sessions[self._id] = self
        adr = self.endpointAddresses().split()
        print "session %s: started. listening on %s:%s" % (self._id, adr[0], ','.join(adr[1:]))

    ## Properties that depend on the state of the media streams
    def getit(self): return min([x.idleTime for x in self.mediaStreams] or [0])
    def getco(self): return [stream.rtpStream.rtpCodec for stream in self.mediaStreams]
    def getty(self): return [stream.type for stream in self.mediaStreams]
    def getta(self): return [self.calledInfo['fromtag'], self.calledInfo['totag']]
    def getby(self):
        bytes = [stream.bytes for stream in self.mediaStreams]
        return reduce(lambda x,y: (x[0]+y[0], x[1]+y[1], x[2]+y[2]), bytes)
    def getpa(self):
        packs = [stream.packets for stream in self.mediaStreams]
        return reduce(lambda x,y: (x[0]+y[0], x[1]+y[1], x[2]+y[2]), packs)

    ## Properties related to call duration
    def getdu(self): return int(round(time.time()) - self.startTime)
    def getsu(self):
        t = self.conversationStart or round(time.time())
        return int(t - self.startTime)
    def getcd(self):
        if self.conversationStart is None:
            return 0
        callDuration = int(round(time.time()) - self.conversationStart)
        if self.didTimeout:
            callDuration = max(callDuration - int(self.idleTime), 0)
        return callDuration

    ## State related properties
    def getdt(self): return (self.idleTime >= self.timeout)
    def getsl(self):
        return 'session %s %s %s %s %s %s' % \
               (self._id, self.callerInfo['from'], self.callerInfo['to'],
                self.encodedAgents[0], self.encodedAgents[1], self.duration)
    def getin(self):
        return {'id': self._id,
                'startTime': self.startTime,
                'conversationStart': self.conversationStart,
                'duration': self.duration,
                'callDuration': self.callDuration,
                'bytes': self.bytes,
                'packets': self.packets,
                'codecs': self.codecs,
                'streamTypes': self.streamTypes,
                'userAgents': self.userAgents,
                'tags': self.tags,
                'didTimeout': self.didTimeout,
                'forceClosed': self.forceClosed}

    codecs       = property(getco)
    streamTypes  = property(getty)
    tags         = property(getta)
    bytes        = property(getby)
    packets      = property(getpa)

    idleTime     = property(getit)
    callDuration = property(getcd)
    duration     = property(getdu)
    setupTime    = property(getsu)

    didTimeout   = property(getdt)
    statline     = property(getsl)
    info         = property(getin)
    
    del(getit, getdt, getco, getty, getta, getdu)
    del(getsu, getby, getpa, getcd, getsl, getin)

    def parseInfo(self, data):
        try:
            entries = data.split('=', 1)[1].split(',')
        except IndexError:
            entries = []
        info = {'from': '', 'to': '', 'fromtag': '', 'totag': ''}
        for entry in entries:
            try:
                key, value = entry.split(':', 1)
            except ValueError:
                key, value = entry, 1
            info[key] = value
        if not info['from']:
            # remove at a later time
            print "warning: SER is using an old mediaproxy.so module. " + \
                  "Please upgrade it else accounting won't work."
        return info

    def updateStreams(self, sessionData):
        streams = [s.split(':') for s in sessionData[2].split(',')]
        streamCount = len(self.mediaStreams)
        if streamCount >= len(streams):
            return
        visibleIP = sessionData[3]
        asymmetric = self.callerInfo.get('asymmetric', 0)
        for sinfo in streams[streamCount:]:
            try:
                mediatype = sinfo[2]
            except IndexError:
                mediatype = 'Audio'
                # remove at a later time
                print "warning: SER is using an old mediaproxy.so module. " + \
                      "Please upgrade it to get correct media type information."
            try:
                stream = MediaStream(self, mediatype)
            except MediaStreamError:
                # check this. also at session creation
                #for stream in self.mediaStreams:
                    #stream.close()
                raise RTPSessionError, "cannot find enough free ports for the media streams"
            stream.setCaller(sinfo, visibleIP, asymmetric)
            self.mediaStreams.append(stream)
        self.oldCount = streamCount
        self.needStreamUpdate = 1
        
    def fillInCalledParty(self, sessionData):
        if not self.complete:
            visibleIP = sessionData[3]
            self.userAgents[1] = sessionData[8].decode('quopri')
            self.encodedAgents[1] = sessionData[8]
            self.calledInfo = self.parseInfo(sessionData[9])
            asymmetric = self.calledInfo.get('asymmetric', 0)
            streams = [s.split(':') for s in sessionData[2].split(',')]
            for i in range(min(len(streams), len(self.mediaStreams))):
                # Should we also re-set the mediatypes here? can the callee change them?
                self.mediaStreams[i].setCalled(streams[i], visibleIP, asymmetric)
            self.conversationStart = round(time.time())
            self.complete = 1
        if self.needStreamUpdate:
            visibleIP = sessionData[3]
            asymmetric = self.calledInfo.get('asymmetric', 0)
            streams = [s.split(':') for s in sessionData[2].split(',')]
            streamCount = len(self.mediaStreams)
            newCount = len(streams)
            for i in range(self.oldCount, min(newCount, streamCount)):
                # Should we also re-set the mediatypes here? can the callee change them?
                self.mediaStreams[i].setCalled(streams[i], visibleIP, asymmetric)
            del(self.oldCount)
            self.needStreamUpdate = 0

    def close(self):
        for stream in self.mediaStreams:
            stream.close()
    __del__ = close  ## Free sockets if a session is deleted without being ended first

    def end(self):
        if self.mediaStreams:
            global Sessions

            self.close()
            try:
                del(Sessions[self._id])
            except KeyError:
                pass

            ## Show summary in syslog
            print("session %s: %s/%s/%s packets, %s/%s/%s bytes (caller/called/relayed)" %
                  ((self._id,) + self.packets + self.bytes))

            ## Do accounting if available
            if radiusdb:
                try:
                    radiusdb.log(self.info)
                except RuntimeError, msg:
                    print "log error: %s" % msg

            what = "ended"
            if self.didTimeout:
                what += " (did timeout)"
            elif self.forceClosed:
                what += " (force closed)"
            print "session %s: %s." % (self._id, what)
            self.mediaStreams = []

    def endpointAddresses(self):
        '''Return the IP address and the RTP data port(s) that will be used by this session'''
        ports = [str(stream.rtpStream.addr[1]) for stream in self.mediaStreams]
        return " ".join([proxyIP] + ports)

