
'''
 Implementation of a mediaproxy request dispatcher for SER.

 The dispatcher implements an architecture where SER will work with multiple
mediaproxy servers (local or remote). The dispatcher is implemented around the
concept of choosing which mediaproxy server to use based on the domain of the
sender or destination. If no mediaproxy server is defined for the given domain
it will fallback to a default mediaproxy server (if configured to use one).

 Mediaproxy servers for a domain are defined using DNS SRV records.
Multiple mediaproxy servers can be defined for a domain by using the priority
and weight attributes of the SRV records. This permits both load balancing the
traffic between multiple mediaproxy servers (if they are defined at the same
priority but with different weights) or mediaproxy server fallbacks (if none
of the servers defined at a priority responds, it will fallback to servers
defined at a lower priority).


Copyright 2004-2005 Dan Pascu
'''

from SocketServer import ThreadingUnixStreamServer, BaseRequestHandler
import socket
import sys, os, grp, atexit, errno
import time
import threading
import random
import cPickle
from utilities import *
from datatypes import *


##
## dispatcher configuration
##
class DispatcherConfig:
    socket        = '/var/run/proxydispatcher.sock'
    group         = 'ser'
    defaultProxy  = '/var/run/mediaproxy.sock'
    timeout       = 24*60*60 ## remove sessions older than 24 hours
    checkInterval = 600      ## check this often for expired sessions

dataTypes = {
    'defaultProxy': MediaProxyAddress
}

## We use this to overwrite some of the settings above on a local basis if needed
from configuration import readSettings
readSettings('Dispatcher', DispatcherConfig, dataTypes)

## Internal state of dispatcher (DON'T TOUCH!!!)

Sessions = {}

## Sessions above should only be accessed after aquiring this lock
sessionLock = threading.RLock()

maxDataSize = 8192

## Fix the environment

## Find where the dispatcher is located
##
## Add path to foreign python modules that we include with mediaproxy for
## convenience since they are not standard. Some are also needed for python < 2.3
Path0 = os.path.normpath((sys.path[0] or '.'))
Path1 = os.path.normpath((sys.path[1] or '.'))
if os.path.exists("%s/dispatcher.py" % Path1):
    ModulesPath = Path1
    sys.path[2:2] = ["%s/%s" % (ModulesPath,  'external')]
else:
    ModulesPath = Path0
    sys.path[1:1] = ["%s/%s" % (ModulesPath,  'external')]

import DNS
DNS.DiscoverNameServers()

## check if we have python 2.3 or newer. Else use some 3rd party module for timeout
## to work with sockets
try:
    socket.setdefaulttimeout
except:
    import timeoutsocket
    socket.setdefaulttimeout = timeoutsocket.setDefaultSocketTimeout
    socket.getdefaulttimeout = timeoutsocket.getDefaultSocketTimeout
    socket.timeout = timeoutsocket.Timeout
    del(timeoutsocket)

##
## Main loop (convenience function)
##

def run(server):
    timer = SessionTimer(DispatcherConfig.checkInterval)
    timer.start()
    server.serve_forever()

def saveSessions():
    sessionLock.acquire()
    try:
        if Sessions:
            sessionsFile = '%s/%s' % (ModulesPath, 'sessions.dat')
            try:
                f = open(sessionsFile, 'w')
            except:
                pass
            else:
                failedDump = False
                try:
                    try:
                        cPickle.dump(Sessions, f)
                    except:
                        failedDump = True
                finally:
                    f.close()
                if failedDump:
                    try:    os.unlink(sessionsFile)
                    except: pass
    finally:
        sessionLock.release()

def restoreSessions():
    sessionLock.acquire()
    try:
        sessionsFile = '%s/%s' % (ModulesPath, 'sessions.dat')
        try:
            f = open(sessionsFile, 'r')
        except:
            pass
        else:
            global Sessions
            try:
                Sessions = cPickle.load(f)
            except:
                pass
            f.close()
            try:    os.unlink(sessionsFile)
            except: pass
    finally:
        sessionLock.release()

## Helper functions

def addDispatcherFlag(data):
    try:
        info = [x for x in data.split('=', 1)[1].split(',') if x]
    except IndexError:
        info = []
    info.append('dispatcher')
    return "info=%s" % ','.join(info)

## Accessing the sessions thread safe

def getSession(sessionId):
    sessionLock.acquire()
    try:
        return Sessions[sessionId]
    finally:
        sessionLock.release()

def allSessions():
    sessionLock.acquire()
    try:
        return Sessions.values()
    finally:
        sessionLock.release()

## Classes

class CommandError(Exception):    pass
class DispatcherError(Exception): pass


class SessionTimer:
    def __init__(self, period):
        self.__period = period
        self.__started = False
        self.__lock = threading.RLock()
        self.__thread = None
    def start(self):
        self.__lock.acquire()
        try:
            assert not self.__started, "timer already started"
            self.__started = True
            self.__run()
        finally:
            self.__lock.release()
    def cancel(self):
        self.__lock.acquire()
        try:
            if self.__thread:
                self.__thread.cancel()
                self.__thread = None
            self.__started = False
        finally:
            self.__lock.release()
    def __run(self):
        ## Check sessions and timeout staled ones
        count = 0
        now = time.time()
        sessionLock.acquire()
        try:
            for session in Sessions.values():
                if now >= session.expire:
                    try:
                        del Sessions[session._id]
                    except KeyError:
                        pass
                    else:
                        count += 1
        finally:
            sessionLock.release()
        if count > 0:
            info("expired %d staled session(s)" % count)
        self.__lock.acquire()
        try:
            if self.__started:
                timer = threading.Timer(self.__period, self.__run)
                timer.setName('SessionTimer')
                timer.setDaemon(True)
                timer.start()
                self.__thread = timer
            else:
                self.__thread = None
        finally:
            self.__lock.release()


#
# Implement Request class as well as a SessionInfo class
# Can probably reuse the existing SessionInfo and adapt it.
#
#class Request(object):
    #commands = {'request': {'minArguments': 9},
                #'lookup':  {'minArguments': 9},
                #'delete':  {'minArguments': 1}}
    #def __init__(self, command):
        #data = command.split()


class SERCommandHandler(ThreadingUnixStreamServer):
    def __init__(self, path=None, group=None):
        ## If we would have passed these as default values they would have been
        ## locked to the values they had when the dispatcher module is loaded
        ## That way we can change the default values after dispatcher is loaded
        ## and we get the updated values
        path = path or DispatcherConfig.socket
        group = group or DispatcherConfig.group
        self.path = path
        self.group = group
        try:
            os.unlink(path)
        except OSError:
            pass
        try:
            ThreadingUnixStreamServer.__init__(self, path, SERRequestHandler)
        except socket.error, why:
            raise DispatcherError, "couldn't create server (%s)" % why[1]
        atexit.register(self.handle_close)

    def server_bind(self):
        ThreadingUnixStreamServer.server_bind(self)
        path = self.path
        try:
            groupId = grp.getgrnam(self.group)[2]
        except KeyError, IndexError:
            ## Make it world writable
            try:
                os.chmod(path, 0777)
            except OSError:
                warning("couldn't set access rights for %s." % path)
                warning("SER may not be able to communicate with us!")
        else:
            ## Make it writable only to the SIP proxy group members
            try:
                os.chown(path, -1, groupId)
                os.chmod(path, 0770)
                #os.chmod(path, S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IWGRP|S_IXGRP)
            except OSError:
                warning("couldn't set access rights for %s." % path)
                warning("SER may not be able to communicate with us!")

    def handle_close(self):
        self.socket.close()
        try:
            os.unlink(self.path)
        except OSError:
            pass


class SERRequestHandler(BaseRequestHandler):
    '''Handles a command received through the control socket'''
    commandMap = {'request': {'minArguments': 9},
                  'lookup':  {'minArguments': 9},
                  'delete':  {'minArguments': 1}}
    
    def handle(self):
        command = self.request.recv(maxDataSize)
        if command:
            try:
                s = time.time()
                result = self.execute(command)
                print "command execution time: %5.2f ms" % ((time.time()-s)*1000)
                if result is not None:
                    self.request.send(result)
            except CommandError, msg:
                error(msg)
        self.request.close() # this is done by close_request() anyway

    def execute(self, command):
        print "command %s" % command.rstrip()

        ## Split command and check if it is valid
        sessionData = command.split()
        if len(sessionData) < 1:
            raise CommandError, "received empty command '%s'" % command.rstrip()
        cmd = sessionData[0].lower()
        if cmd not in self.commandMap.keys():
            raise CommandError, "received unknown command '%s'" % command.rstrip()

        minArguments = self.commandMap[cmd]['minArguments']
        if len(sessionData) < minArguments + 1:
            suffix = (minArguments!=1 and 's') or ''
            raise CommandError, "the '%s' command needs at least %d argument%s: '%s'" % \
                  (cmd, minArguments, suffix, command.rstrip())

        ## And do the mumbo-jumbo thing
        if cmd == 'request':
            ## Lock before get and release after add to avoid concurency issues
            sessionId = sessionData[1]
            sessionLock.acquire()
            try:
                session = Sessions.get(sessionId, None)
            finally:
                sessionLock.release()
            if session is not None:
                sessionData[9] = addDispatcherFlag(sessionData[9])
                command = '%s\n' % ' '.join(sessionData)
                return session.forward(command)
            fromDomain = sessionData[4]
            toDomain   = sessionData[6]
            localFrom = (sessionData[5]=='local')
            localTo   = (sessionData[7]=='local')
            domain = (localFrom and fromDomain) or (localTo and toDomain) or None
            if domain is not None:
                proxy = MediaProxy('domain://%s' % domain).address
            else:
                proxy = MediaProxy(DispatcherConfig.defaultProxy).address
            sessionLock.acquire()
            try:
                try:
                    session = Sessions[sessionId]
                except KeyError:
                    info = SessionInfo(sessionData[9])
                    if info.totag:
                        ## Don't create a new session on re-INVITEs
                        return None
                    if not proxy:
                        warning("no proxy available for call: %s -> %s" %
                                (fromDomain, toDomain))
                        return None
                    session = MediaSession(sessionId, proxy)
                    Sessions[sessionId] = session
            finally:
                sessionLock.release()
            sessionData[9] = addDispatcherFlag(sessionData[9])
            command = '%s\n' % ' '.join(sessionData)
            return session.forward(command)
        elif cmd == 'lookup':
            sessionId = sessionData[1]
            try:
                session = getSession(sessionId)
            except KeyError:
                #warning("trying to lookup session with non-existent id: '%s'" % sessionId)
                return None
            sessionData[9] = addDispatcherFlag(sessionData[9])
            command = '%s\n' % ' '.join(sessionData)
            return session.forward(command)
        elif cmd == 'delete':
            sessionId = sessionData[1]
            try:
                session = getSession(sessionId)
            except KeyError:
                return None
            if len(sessionData) == 2:
                # Fix for older mediaproxy modules that don't send info.
                # remove in a later version (and increment minArguments to 2 for delete)
                sessionData.append('info=dispatcher')
            else:
                sessionData[2] = addDispatcherFlag(sessionData[2])
            command = '%s\n' % ' '.join(sessionData)
            result = session.forward(command)
            session.release()
            return result
        else:
            raise CommandError, "received unknown command: '%s'" % command.rstrip()


class SessionInfo(dict):
    def __init__(self, data):
        try:
            entries = data.split('=', 1)[1].split(',')
        except IndexError:
            entries = []
        self.update({'from': '', 'to': '', 'fromtag': '', 'totag': ''})
        for entry in entries:
            try:
                key, value = entry.split(':', 1)
            except ValueError:
                key, value = entry, 1
            self[key] = value
        if not self['from']:
            # remove at a later time
            print("warning: SER is using an old mediaproxy.so module. "
                  "Please upgrade it, else accounting won't work.")
    def __setitem__(self, key, value):
        self.__dict__[key] = value
        dict.__setitem__(self, key, value)
    def __delitem__(self, key):
        dict.__delitem__(self, key)
        del self.__dict__[key]
    __setattr__ = __setitem__
    def __delattr__(self, name):
        try:
            del self.__dict__[name]
        except KeyError:
            raise AttributeError, "'%s' object has not attribute '%s'" % (self.__class__.__name__, name)
        else:
            dict.__delitem__(self, name)
    def update(self, other):
        dict.update(self, other)
        for key, value in other.items():
            self.__dict__[key] = value


class MediaSession(object):
    def __init__(self, sessionId, address):
        self._id = sessionId
        self.address = address
        self.sessionStart = time.time()
        self.expire = self.sessionStart + DispatcherConfig.timeout
    def getst(self):
        if type(self.address) is tuple:
            return socket.AF_INET
        else:
            return socket.AF_UNIX
    def getas(self):
        if type(self.address) is tuple:
            return "%s:%s" % self.address
        else:
            return self.address
    socktype = property(getst)
    address_str = property(getas)
    del getst, getas
    
    def forward(self, command):
        if self.address is None:
            return None
        s = socket.socket(self.socktype, socket.SOCK_STREAM)
        try:
            s.settimeout(5)
        except AttributeError:
            pass
        result = None
        try:
            try:
                s.connect(self.address)
                s.send(command)
                result = s.recv(maxDataSize)
            except (socket.error, socket.gaierror, socket.herror, socket.timeout), why:
                try:               motive = why[1]
                except IndexError: motive = why
                error("couldn't forward command to %s (%s)" % (self.address_str, motive))
                return None
        finally:
            s.close()
        print "forwarding to mediaproxy on %s: got: '%s'" % (self.address_str, result.rstrip())
        return result

    def release(self):
        sessionLock.acquire()
        try:
            try:
                del(Sessions[self._id])
            except KeyError:
                pass
        finally:
            sessionLock.release()


class MediaProxy(object):
    '''
    Class that defines a single MediaProxy server or a collection of
    MediaProxy servers as defined by a domain (using SRV records)
    '''
    def __init__(self, url):
        try:
            url + ''
        except TypeError:
            if url is None or isinstance(url, tuple):
                self.address = url
                return
            else:
                raise TypeError, "invalid MediaProxy address type"
        domainId = 'domain://'
        if url.startswith(domainId):
            domain = url[len(domainId):]
            mediaproxyrec = "_mediaproxy._tcp.%s" % domain
            try:
                dns = DNS.DnsRequest()
                answers = dns.req(name=mediaproxyrec, qtype='SRV', timeout=7).answers
            except (DNS.Base.DNSError, socket.error), why:
                error("couldn't retrieve DNS SRV record '%s': %s" % (mediaproxyrec, why))
                self.address = None
                return
            if not answers:
                self.address = MediaProxy(DispatcherConfig.defaultProxy).address
                return
            servers = {}
            for x in answers:
                try:
                    (priority, weight, port, address) = x['data']
                except:
                    continue
                array = servers.setdefault(priority, [])
                array.append((weight, address, port))
            priorities = servers.keys()
            priorities.sort()
            allproxies = []
            for p in priorities:
                all = servers[p] ## all servers at the same priority
                if len(all) == 1:
                    allproxies.append((all[0][1:]))
                else:
                    weights = [x[0] or 1 for x in all]
                    try:
                        totalWeight = sum(weights)
                    except NameError:
                        ## This is for python older than 2.3
                        totalWeight = reduce(lambda x,y: x+y, weights)
                    weights = [int(w*100/totalWeight) for w in weights]
                    array = reduce(lambda x,y: x+y, [weights[i]*[i] for i in
                                                     xrange(len(weights))])
                    while array:
                        index = random.choice(array)
                        allproxies.append(all[index][1:])
                        array = [x for x in array if x!=index]
                    del(array)
            ## Now determine which of the servers are actually active and can be used
            # Start all connections in non-blocking mode the revert to blocking and poll each.
            # That way we only wait for 5 seconds no matter how many, instead of n*5 as it is now.
            # Doesn't work, as we will not be able to use timeouts after reverting to blocking
            # mode. ev start first in blocking mode and the rest in nonblocking mode and check
            # them only if first did timeout.
            for proxy in allproxies:
                s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                s.settimeout(3)
                try:
                    try:
                        s.connect(proxy)
                    except (socket.error, socket.gaierror, socket.herror, socket.timeout):
                        continue
                    else:
                        s.settimeout(1)
                        try:
                            s.send('version\n')
                            res = s.recv(1024)
                        except (socket.error, socket.gaierror, socket.herror, socket.timeout):
                            continue
                        if res:
                            break
                finally:
                    s.close()
            else:
                warning("none of the proxies for domain %s have responded" % domain)
                proxy = None
            self.address = proxy
        else:
            self.address = MediaProxyAddress(url) # implement this
    def __setattr__(self, name, value):
        readonly = ('address')
        if name in readonly and self.__dict__.has_key(name):
            raise AttributeError, "'%s' is a readonly attribute" % name
        self.__dict__[name] = value

